import 'dart:async';
import 'dart:collection';

import '../error/errors.dart';
import 'component.dart';
import 'providers.dart';

class MiodiConfiguration {
  final HashMap<String, HashMap<Type, Provider>> _configuration = HashMap();
  final HashSet<String> _namespaces = HashSet();
  ScopeConfiguration _scopeConfiguration;

  MiodiConfiguration(List<MiodiComponent> components) {
    components.forEach((component) {
      var componentNamespace = component.componentNamespace;
      _namespaces.add(componentNamespace);
      if (_configuration.containsKey(componentNamespace))
        _configuration[componentNamespace]
            .addAll(component.componentConfiguration);
      else
        _configuration[componentNamespace] = component.componentConfiguration;
    });
    _scopeConfiguration = ScopeConfiguration(components, _namespaces);
  }

  ScopeConfiguration get scopeConfiguration => _scopeConfiguration;

  T provide<T>([String componentNamespace]) {
    T obj;
    if (componentNamespace == null) {
      for (String namespace in _namespaces) {
        if (_configuration[namespace].containsKey(T)) {
          obj = _configuration[namespace][T].provide() as T;
          break;
        }
      }
    } else {
      if (!_configuration.containsKey(componentNamespace))
        throw BadComponentNamespaceException(componentNamespace);
      if (!_configuration[componentNamespace].containsKey(T))
        throw NotFoundProviderException(T, componentNamespace);
      obj = _configuration[componentNamespace][T].provide() as T;
    }
    if (obj == null) throw NotFoundProviderException(T);
    return obj;
  }

  Future<T> provideAsync<T>([String componentNamespace]) async {
    return provide<T>(componentNamespace);
  }
}

class ScopeConfiguration {
  final HashMap<String, HashMap<String, Map<Type, ScopeProvider>>>
      _scopeConfiguration = HashMap();
  final Map<String, HashSet<String>> _allScopes = HashMap();
  final HashSet<String> _namespaces;

  ScopeConfiguration(List<MiodiComponent> components, this._namespaces) {
    components
        .where((component) => component.scopeConfiguration.isNotEmpty)
        .forEach((component) {
      var componentNamespace = component.componentNamespace;
      if (_scopeConfiguration.containsKey(componentNamespace)) {
        _scopeConfiguration[componentNamespace]
            .addAll(component.scopeConfiguration);
      } else {
        _scopeConfiguration[componentNamespace] = component.scopeConfiguration;
      }
      _addScopes(componentNamespace, component.scopeConfiguration.keys);
    });
  }

  void _addScopes(String componentNamespace, Iterable<String> keys) {
    if (!_allScopes.containsKey(componentNamespace))
      _allScopes[componentNamespace] = HashSet();
    _allScopes[componentNamespace].addAll(keys);
  }

  T provide<T>(String scope, [String componentNamespace]) {
    T model;
    if (componentNamespace == null) {
      for (String namespace in _namespaces) {
        if (_scopeConfiguration[namespace].containsKey(scope)) {
          if (_scopeConfiguration[namespace][scope].containsKey(T)) {
            model = _scopeConfiguration[namespace][scope][T].provide() as T;
            break;
          }
        }
      }
    } else {
      if (!_scopeConfiguration.containsKey(componentNamespace))
        throw BadComponentNamespaceException(componentNamespace);
      if (!_scopeConfiguration[componentNamespace][scope].containsKey(T))
        throw NotFoundProviderException(T, componentNamespace);
      if (!_scopeConfiguration[componentNamespace].containsKey(scope))
        throw BadScopeNameException(scope);
      model = _scopeConfiguration[componentNamespace][scope][T].provide() as T;
    }
    if (model == null) throw NotFoundProviderException(T);
    return model;
  }

  Future<T> provideAsync<T>(String scope, [String componentNamespace]) async {
    return provide(scope, componentNamespace);
  }

  void clearScope(String scope, [String componentNamespace]) {
    if (componentNamespace != null) {
      if (!_scopeConfiguration.containsKey(componentNamespace))
        throw BadComponentNamespaceException(componentNamespace);
      if (!_scopeConfiguration[componentNamespace].containsKey(scope))
        throw BadScopeNameException(scope);
      _scopeConfiguration[componentNamespace][scope].forEach((_, p) {
        p.clear();
      });
    } else {
      _scopeConfiguration.forEach((namespace, map) {
        if (map.containsKey(scope))
          map[scope].forEach((_, scopeProvider) {
            scopeProvider.clear();
          });
      });
    }
  }

  void clearScopes(Iterable<String> scopes) {
    scopes.forEach(clearScope);
  }
  void clearScopesWithNamespace(Map<String, Iterable<String>> inactiveScopes) {
    inactiveScopes.forEach((namespace, scopes) {
      scopes.forEach((scope) {clearScope(scope, namespace);});
    });
  }

  void clearAllScope([String componentNamespace]) {
    if (componentNamespace == null)
      _scopeConfiguration.forEach((_, mapScope) {
        mapScope.forEach((_, mapProvider) {
          mapProvider.forEach((_, provider) {
            provider.clear();
          });
        });
      });
    else {
      if (!_scopeConfiguration.containsKey(componentNamespace)) return;
      _scopeConfiguration[componentNamespace].forEach((_, mapProvider) {
        mapProvider.forEach((_, provider) {
          provider.clear();
        });
      });
    }
  }

  void clearInactiveScopesWithNamespace(Map<String, Iterable<String>> activeScopes) {
    _allScopes.forEach((namespace, scopes) {
      if (!activeScopes.containsKey(namespace))
        clearAllScope(namespace);
      else {
        scopes.where((scope) => !activeScopes[namespace].contains(scope))
            .forEach((inactiveScope) {
              clearScope(inactiveScope, namespace);
        });
      }
    });
  }

  void clearInactiveScopes(Iterable<String> activeScopes) {
    _allScopes.forEach((namespace, scopes) {
      scopes.where((s) => !activeScopes.contains(s)).forEach((scope) {
        clearScope(scope, namespace);
      });
    });
  }
}
